@@ -39,9 +39,12 @@ from transformers import (
     set_seed,
 )
 from transformers import is_tf_available
-from transformers.trainer_utils import PredictionOutput
 from mc_transformers.utils_mc import MultipleChoiceDataset, Split, processors
-from mc_transformers.data_classes import WindowPrediction
+from mc_transformers.data_classes import (
+    WindowPrediction,
+    PredictionOutputWithIds,
+    DataCollatorWithIds,
+)
 
 
 if is_tf_available():
@@ -60,6 +63,11 @@ def simple_accuracy(preds, labels):
     return (preds == labels).mean()
 
 
+def compute_metrics(p: EvalPrediction) -> Dict:
+    preds = np.argmax(p.predictions, axis=1)
+    return {"acc": simple_accuracy(preds, p.label_ids)}
+
+
 def softmax(preds, axis=None):
     # Taken from: https://nolanbconaway.github.io/blog/2017/softmax-numpy.html
     if axis is None:
@@ -348,24 +356,40 @@ def save_predictions(processor, results, args, split):
 def save_results(processor, results, args, split):
     # only predict method returns prediction outputs,
     # evaluate and train only return the metrics
-    if isinstance(results, PredictionOutput):
+    if isinstance(results, PredictionOutputWithIds):
         save_metrics(results.metrics, args, split)
         save_predictions(processor, results, args, split)
     else:
         save_metrics(results, args, split)
 
 
-def main():
-    # See all possible arguments in src/transformers/training_args.py
-    # or by passing the --help flag to this script.
-    # We now keep distinct sets of args, for a cleaner separation of concerns.
+def pair_predictions_with_ids(results, data_collator):
+    return PredictionOutputWithIds(
+        predictions=results.predictions,
+        label_ids=results.label_ids,
+        example_ids=data_collator.example_ids,
+        metrics=results.metrics,
+    )
+
+
+def setup(argc=None, **kwargs):
+    if argc is None:
+        argc = sys.argv[1:]
     parser = HfArgumentParser((
         ModelArguments, DataTrainingArguments,
         DirArguments, TrainingArguments, WindowArguments
     ))
-    if len(sys.argv) > 1 and sys.argv[1].endswith('.json'):
+    if (
+        isinstance(argc, list) and
+        len(argc) == 1 and
+        argc[0].endswith('.json')
+    ):
         model_args, data_args, dir_args, training_args, window_args = (
-            parser.parse_json_file(sys.argv[1])
+            parser.parse_json_file(argc[0])
+        )
+    elif isinstance(argc, dict):
+        model_args, data_args, dir_args, training_args, window_args = (
+            parser.parse_dict(argc)
         )
     else:
         model_args, data_args, dir_args, training_args, window_args = (
@@ -415,30 +439,39 @@ def main():
     except KeyError:
         raise ValueError("Task not found: %s" % (data_args.task_name))
 
-    # Load pretrained model and tokenizer
-    #
-    # Distributed training:
-    # The .from_pretrained methods guarantee that only one local process can concurrently
-    # download model & vocab.
+    config_kwargs = kwargs.pop('config_kwargs', {})
+    tokenizer_kwargs = kwargs.pop('tokenizer_kwargs', {})
+    model_kwargs = kwargs.pop('model_kwargs', {})
 
     config = AutoConfig.from_pretrained(
         model_args.config_name if model_args.config_name else model_args.model_name_or_path,
         num_labels=num_labels,
         finetuning_task=data_args.task_name,
         cache_dir=model_args.cache_dir,
+        **config_kwargs,
     )
     tokenizer = AutoTokenizer.from_pretrained(
         model_args.tokenizer_name if model_args.tokenizer_name else model_args.model_name_or_path,
         cache_dir=model_args.cache_dir,
+        **tokenizer_kwargs,
     )
     model = AutoModelForMultipleChoice.from_pretrained(
         model_args.model_name_or_path,
         from_tf=bool(".ckpt" in model_args.model_name_or_path),
         config=config,
         cache_dir=model_args.cache_dir,
+        **model_kwargs,
+    )
+
+    return all_args, processor, config, tokenizer, model
+
+
+def main():
+    all_args, processor, config, tokenizer, model = setup()
+    model_args, data_args, dir_args, training_args, window_args = (
+        all_args.values()
     )
 
-    # Get datasets
     train_dataset = (
         MultipleChoiceDataset(
             data_dir=data_args.data_dir,
@@ -486,9 +519,7 @@ def main():
         else None
     )
 
-    def compute_metrics(p: EvalPrediction) -> Dict:
-        preds = np.argmax(p.predictions, axis=1)
-        return {"acc": simple_accuracy(preds, p.label_ids)}
+    data_collator = DataCollatorWithIds()
 
     # Initialize our Trainer
     trainer = Trainer(
@@ -497,6 +528,7 @@ def main():
         train_dataset=train_dataset,
         eval_dataset=eval_dataset,
         compute_metrics=compute_metrics,
+        data_collator=data_collator.collate,
     )
 
     # Training
@@ -518,28 +550,34 @@ def main():
         logger.info("*** Evaluate (train set)***")
         result = trainer.predict(train_dataset)
         if trainer.is_world_master():
+            result = pair_predictions_with_ids(result, data_collator)
             save_results(
                 processor, result, all_args, split=Split.train
             )
             results['train'] = result
+            data_collator.drop_ids()
 
     if training_args.do_eval:
         logger.info("*** Evaluate ***")
         result = trainer.predict(eval_dataset)
         if trainer.is_world_master():
+            result = pair_predictions_with_ids(result, data_collator)
             save_results(
                 processor, result, all_args, split=Split.dev
             )
             results['eval'] = result
+            data_collator.drop_ids()
 
     if training_args.do_predict:
         logger.info("*** Test ***")
         result = trainer.predict(test_dataset)
         if trainer.is_world_master():
+            result = pair_predictions_with_ids(result, data_collator)
             save_results(
                 processor, result, all_args, split=Split.test
             )
             results['test'] = result
+            data_collator.drop_ids()
 
     return results
 
